-- upvalues
local LibProcessable = LibStub("LibProcessable")
local tonumber = tonumber
local tconcat = table.concat
local twipe = table.wipe
local tinsert = table.insert
local gsub = string.gsub
local lower = string.lower
local print = print

-- local variables
local disenchantBuffName = GetSpellInfo(161724) --Essence Attunement
local bagIDs = {0, 1, 2, 3, 4}
local ignoredCache = {}
local garrisonMaps = {
	[1152] = true, -- FW Horde Garrison Level 1
	[1330] = true, -- FW Horde Garrison Level 2
	[1153] = true, -- FW Horde Garrison Level 3
	[1154] = true, -- FW Horde Garrison Level 4
	[1158] = true, -- SMV Alliance Garrison Level 1
	[1331] = true, -- SMV Alliance Garrison Level 2
	[1159] = true, -- SMV Alliance Garrison Level 3
	[1160] = true, -- SMV Alliance Garrison Level 4
}

-- settings
local Defaults = {
	["ignoreItemSets"] = true,		-- filters out items that are part of any of your equipmentsets. They are not added to the ignore list.
	["positionLocked"] = false,		-- controls whether or not the button is movable by dragging.
}

-- localization
BINDING_HEADER_GARRISONDISENCHANT = "Garrison Disenchant"
_G["BINDING_NAME_CLICK GarrisonDisenchant:LeftButton"] = "Disenchant Item Currently Shown"

-- frame creation
local GD = CreateFrame("Button", "GarrisonDisenchant", UIParent, "SecureActionButtonTemplate, ActionButtonTemplate")
GD:SetPoint("CENTER", 0, 100)
GD:SetWidth(60)
GD:SetHeight(60)
GD:Hide()
GD:SetMovable(true)
GD:RegisterForClicks("LeftButtonUp", "RightButtonUp")
GD.Version = GetAddOnMetadata("GarrisonDisenchant", "Version")

GD.Tex = GD:CreateTexture(nil, "ARTWORK")
GD.Tex:SetAllPoints()

GD.HotKey = GD:CreateFontString(nil, "OVERLAY", "SystemFont_Outline")
GD.HotKey:SetShadowColor(0,0,0)
GD.HotKey:SetShadowOffset(1,-1)
GD.HotKey:SetPoint("CENTER")

-- script handlers
GD:SetScript("OnDragStart", GD.StartMoving)
GD:SetScript("OnDragStop", GD.StopMovingOrSizing)
GD:SetScript("OnLeave", function(self) GameTooltip:Hide() end)
GD:SetScript("OnEnter", function(self)
	GameTooltip:SetOwner(self, "ANCHOR_CURSOR");
	GameTooltip:SetHyperlink(self.itemLink)
	GameTooltip:Show()
end)
GD:SetScript("OnHide", function(self)
	if self.isMoving then
		self:StopMovingOrSizing();
		self.isMoving = false;
	end
end)

-- functions
local function GDPrint(msg)
	print('|cff00b3ffGD:|r', msg)
end

local function BuildIgnoredCache()
	twipe(ignoredCache)
	for itemID in pairs(GarrisonDisenchantDB["Ignored"]) do
		if not ignoredCache[itemID] then ignoredCache[itemID] = true end
	end
end

local function UpdateButton(bagID, slot)
	if InCombatLockdown() then return; end

	if (not bagID and not slot) then
		GD.itemID = nil
		GD.itemLink = nil
		GD:Hide()
	else
		local texture = GetContainerItemInfo(bagID, slot)
		local itemLink = GetContainerItemLink(bagID, slot)
		local itemID = GetContainerItemID(bagID, slot)
		GD.Tex:SetTexture(texture)
		GD.itemLink = itemLink
		GD.itemID = itemID
		GD:SetAttribute("type1", "macro")
		GD:SetAttribute("macrotext1", "/use "..bagID.. " "..slot)
		GD:Show()
	end
end

local function FindItem()
	for _, bagID in pairs(bagIDs) do
		local numSlots = GetContainerNumSlots(bagID)
		for slot = 1, numSlots do
			local itemID = GetContainerItemID(bagID, slot)

			if itemID then
				local isDisenchantable = LibProcessable:IsDisenchantable(itemID, nil, true)
				local isInSet = GetContainerItemEquipmentSetInfo(bagID, slot)

				if isDisenchantable and not ignoredCache[itemID] and ((GarrisonDisenchantDB.Settings.ignoreItemSets and not isInSet) or not GarrisonDisenchantDB.Settings.ignoreItemSets) then
					UpdateButton(bagID, slot)
					return
				end
			end
		end
	end
end

local function UpdateBinding()
	local text = GetBindingKey("CLICK GarrisonDisenchant:LeftButton")
	if text then
		text = gsub(text, 'SHIFT', "Shift")
		text = gsub(text, 'ALT', "Alt")
		text = gsub(text, 'CTRL', "Ctrl")
		text = gsub(text, 'BUTTON', "M")
		text = gsub(text, 'MOUSEWHEELUP', "MwU")
		text = gsub(text, 'MOUSEWHEELDOWN', "MwD")
		text = gsub(text, 'NUMPAD', "N")
		text = gsub(text, 'PAGEUP', "PU")
		text = gsub(text, 'PAGEDOWN', "PD")
		text = gsub(text, 'SPACE', "SpB")
		text = gsub(text, 'INSERT', "Ins")
		text = gsub(text, 'HOME', "Hm")
		text = gsub(text, 'DELETE', "Del")
		text = gsub(text, 'NMULTIPLY', "*")
		text = gsub(text, 'NMINUS', "N-")
		text = gsub(text, 'NPLUS', "N+")
	end
	GD.HotKey:SetText(text);
end

local function GetItemIDFromItem(item)
	if not item then return; end
	local itemID
	if tonumber(item) then
		itemID = item
	else
		local _, itemLink = GetItemInfo(item)
		if itemLink then
			local itemString = string.match(itemLink, "item[%-?%d:]+")
			local _, id = strsplit(":", itemString)
			itemID = id
		end
	end

	return itemID
end

function GD:IgnoreItem(item)
	local itemID
	if not item then
		itemID = GD.itemID
	else
		local itemString = string.match(item, "item[%-?%d:]+")
		if itemString then
			local _, id = strsplit(":", itemString)
			itemID = id
		else
			itemID = GetItemIDFromItem(item) or GD.itemID
		end
	end

	GarrisonDisenchantDB["Ignored"][tonumber(itemID)] = true
	BuildIgnoredCache()
	UpdateButton()
	FindItem()
	GDPrint("Following item was ignored: " .. (GetItemInfo(itemID)))
end

local function RemoveItemFromIgnore(item)
	if item == "all" then
		twipe(ignoredCache)
		twipe(GarrisonDisenchantDB["Ignored"])
		GDPrint("All items removed from ignore.")
	elseif item then
		local itemID
		local itemString = string.match(item, "item[%-?%d:]+")
		if itemString then
			local _, id = strsplit(":", itemString)
			itemID = tonumber(id)
		else
			itemID = tonumber(GetItemIDFromItem(item))
		end
		if not itemID then return; end

		if GarrisonDisenchantDB["Ignored"][itemID] then GarrisonDisenchantDB["Ignored"][itemID] = nil end
		GDPrint("Item removed from ignore: " .. (GetItemInfo(itemID)))
	end
	BuildIgnoredCache()
	FindItem()
end

local function GetIgnoredItemsAsString()
	local temp = {}
	for itemID in pairs(ignoredCache) do
		local itemName = GetItemInfo(itemID)
		local itemNameAndID = itemName .. " ("..itemID..")"
		tinsert(temp, itemNameAndID)
	end
	local itemString = tconcat(temp, ", ")
	if itemString == "" then itemString = "None" end
	return itemString
end

local function UpdateFilterSetting(setting)
	local filterEnabled
	if setting == "on" then
		filterEnabled = true
	elseif setting == "off" then
		filterEnabled = false
	end

	GarrisonDisenchantDB.Settings.ignoreItemSets = filterEnabled
	GDPrint("Equipment Set Filter is now " .. (filterEnabled == true and "ENABLED" or "DISABLED"))
	UpdateButton()
	FindItem()
end

-- slash commands
SLASH_GARRISONDISENCHANT1, SLASH_GARRISONDISENCHANT2 = "/garrisondisenchant", "/gd"
local function handler(arg)
	local command, param = arg:match("^(%S*)%s*(.-)$")
	if command then command = lower(command) end
	if param then param = lower(param) end

	if command == "ignore" then
		if param and param ~= "" then
			GD:IgnoreItem(param)
		end

	elseif command == "remove" then
		if param and param ~= "" then
			RemoveItemFromIgnore(param)
		end

	elseif command == "list" or command == "listignore" then
		GDPrint("List of Ignored Items: " .. GetIgnoredItemsAsString())

	elseif command == "filtersets" then
		if param and param ~= "" and (param == "on" or param == "off") then
			UpdateFilterSetting(param)
		end

	elseif command == "unlock" then
		GD:RegisterForDrag("LeftButton")
		GarrisonDisenchantDB["Settings"].positionLocked = false
		GDPrint("Button Is Now Movable")

	elseif command == "lock" then
		GD:RegisterForDrag()
		GarrisonDisenchantDB["Settings"].positionLocked = true
		GDPrint("Button Position Locked")

	elseif command == "version" then
		GDPrint("Installed Version: " .. GD.Version)

	else
		GDPrint("Usage:")
		GDPrint("/gd ignore [ itemID | itemName | itemLink ]")
		GDPrint("/gd remove [ itemID | itemLink | all ]")
		GDPrint("/gd filtersets [ on | off ]")
		GDPrint("/gd list | listignore")
		GDPrint("/gd unlock")
		GDPrint("/gd lock")
		GDPrint("/gd version")
	end
end
SlashCmdList["GARRISONDISENCHANT"] = handler

-- event handling
GD:RegisterEvent("PLAYER_LOGIN")
GD:SetScript("OnEvent", function(self, event, arg1, arg2, arg3, arg4, arg5)
	if event == "PLAYER_LOGIN" then

		--Initialize saved vars
		if GarrisonDisenchantDB == nil then GarrisonDisenchantDB = {} end
		if GarrisonDisenchantDB.Settings == nil then GarrisonDisenchantDB.Settings = Defaults end
		if GarrisonDisenchantDB.Ignored == nil then GarrisonDisenchantDB.Ignored = {} end

		--Make the button movable if unlocked
		if not GarrisonDisenchantDB["Settings"].positionLocked then
			GD:RegisterForDrag("LeftButton")
		end

		--Register helper events
		self:RegisterEvent("GARRISON_UPDATE")
		self:RegisterEvent("UPDATE_BINDINGS")

		--Build ignore cache and update icon
		UpdateBinding()
		BuildIgnoredCache()
		if UnitBuff("player", disenchantBuffName) then --In case of /reload while having the disenchant buff active
			FindItem()
			self:RegisterEvent("CURRENT_SPELL_CAST_CHANGED")
		end

		--Set right click ignore action
		GD:SetAttribute("type2", "macro")
		GD:SetAttribute("macrotext2", "/run GarrisonDisenchant:IgnoreItem()")
		
		--Print addon information
		GDPrint("GarrisonDisenchant version: " .. GD.Version)
		GDPrint("Use /gd or /garrisondisenchant to configure")

	elseif event == "UPDATE_BINDINGS" then
		UpdateBinding()

	elseif event == "GARRISON_UPDATE" then
		local mapID = select(8, GetInstanceInfo())
		if garrisonMaps[mapID] then
			self:RegisterEvent("SPELL_CONFIRMATION_PROMPT")
			self:RegisterEvent("UNIT_AURA")
			self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
		else
			self:UnregisterEvent("SPELL_CONFIRMATION_PROMPT")
			self:UnregisterEvent("UNIT_AURA")
			self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
			self:UnregisterEvent("CURRENT_SPELL_CAST_CHANGED")
		end

	elseif event == "UNIT_AURA" and arg1 == "player" then
		if not UnitBuff("player", disenchantBuffName) then
			UpdateButton() --Hide button
		end

	elseif event == "CURRENT_SPELL_CAST_CHANGED" then
		if UnitBuff("player", disenchantBuffName) then
			FindItem()
		end

	elseif event == "SPELL_CONFIRMATION_PROMPT" and tonumber(arg1) == 161736 then --Disenchant prompt from clicking the Essence Font
		self:RegisterEvent("CURRENT_SPELL_CAST_CHANGED") --Subsequent clicks on the Essence Font triggers this event

	elseif event == "UNIT_SPELLCAST_SUCCEEDED" and arg1 == "player" and tonumber(arg5) == 160201 then --Disenchant
		self:RegisterEvent("BAG_UPDATE") --Successful disenchant leads to looting and triggers this event

	elseif event == "BAG_UPDATE" then
		self:UnregisterEvent("BAG_UPDATE")
		UpdateButton()
		FindItem() --An item was just disenchanted and looted, find next item
	end
end)